#include ".\actwesolver.h"

ActWESolver::ActWESolver(void)
{
	loaded = false;
	g = NULL;
	dev_cols = dev_rows = dev_diag45 = dev_diag135 = NULL;
	prev_c1 = -100; prev_c2 = -100; 
	iterate_num = 0;
	totol_elapsetime_ms = 0;
}

ActWESolver::~ActWESolver(void)
{
	if(g)
		delete g;
	if(dev_cols){
		cvReleaseMat(&dev_cols);
		cvReleaseMat(&dev_rows);
		cvReleaseMat(&dev_diag135);
		cvReleaseMat(&dev_diag45);
	}
}

void ActWESolver::loadContour(int *x, int *y, int num)
{
	g->add_contour(x, y, num);
}

double ActWESolver::H(double val)
{	
	return (1+2*atan(val/epsilon)/PI)/2.0;
}

double ActWESolver::Delta(double val)
{
	return epsilon / ((epsilon*epsilon + val*val) *PI) ;
}

void ActWESolver::loadParam(CvMat *varpi_i, CvMat *u0_i, double mu_i, double v_i, 
			   double lambda1_i, double lambda2_i,
			   double p_i, double epsilon_i, double dt_i, double k_i, bool video, char *videofname = "lsout.avi") 
{
	mu = mu_i;
	v = v_i;
	lambda1 = lambda1_i;
	lambda2 = lambda2_i;
	p = p_i;
	dt = dt_i;
	k = k_i;
	epsilon = epsilon_i;
	varpi = cvCloneMat(varpi_i);
	u0 = cvCloneMat(u0_i);
	
	width = varpi->height;
	height = varpi->width;

	if(video){
		vout= cvCreateVideoWriter(videofname, 541215044, 18, cvSize(width, height));
	}else{
		vout = NULL;
	}

	
	dx = 1;//	/width
	dy = 1;//	/height
	loaded = true;

	//  Graph cuts
	g = new TGraphImage(height, width);
	
	dev_rows = cvCreateMat(width, height, CV_32FC1);
	dev_cols = cvCreateMat(width, height, CV_32FC1);
	dev_diag45 = cvCreateMat(width, height, CV_32FC1);
	dev_diag135 = cvCreateMat(width, height, CV_32FC1);
	cvSetZero(dev_rows);
	cvSetZero(dev_cols);
	cvSetZero(dev_diag45);
	cvSetZero(dev_diag135);

	for(int j=0; j<(width-1); j++){
		for(int i=0; i<(height-1); i++){
			g -> add_edge(i, j, i, j+1,mu);
			g -> add_edge(i, j, i+1, j,mu);
		}
	}
	for(int i = 0; i < height - 1; i++) {
		g->add_edge(i, width - 1, i+1, width - 1, mu);
	}
	for(int j = 0; j < width - 1; j++) {
		g->add_edge(height - 1, j, height - 1, j+1, mu);
	}
}

/*	
		Calculate the Energy function iteratively

*/
void ActWESolver::Iterate()
{
	double val, val1;
	int x,y; //for saturation
	CvMat *tmpmat = cvCloneMat(varpi);

	//Performance evluation
	struct __timeb64 tstruct1,tstruct2;
	_time64( &timebuffer_before);
	_ftime64(&tstruct1);

	calC();	// calculate c1 and c2

	double C1, C2, C3, C4; //coefficient of \varpi_{i+1,j}^{n+1}
	for(int i=0; i<width; i++){
		for(int j=0; j<height; j++){
			val1 = Delta(cvmGet(varpi, i,j)); 
			val = 0;
			C1 = 1.0/DenoX(i,j);
			C2 = 1.0/DenoX(i-1, j);
			C3 = 1.0/DenoY(i,j);
			C4 = 1.0/DenoY(i,j-1);
			x = i+1; y = j; Saturate(x,y);
			val += C1*cvmGet(varpi, x,y);		// n
			x = i-1; y = j; Saturate(x,y);
			val += C2*cvmGet(tmpmat, x,y); // n+1
			x = i; y = j+1; Saturate(x,y);
			val += C3*cvmGet(varpi, x,y);		//n
			x = i; y = j-1; Saturate(x,y);
			val += C4*cvmGet(tmpmat, x,y);		//n+1

			val *= mu/(dx*dy);

			val += - v - lambda1 * (cvmGet(u0,i,j)-c1)* (cvmGet(u0,i,j)-c1)
					   + lambda2 * (cvmGet(u0,i,j)-c2)*(cvmGet(u0,i,j)-c2);

			val *= dt * val1;
			val += cvmGet(varpi, i,j);
			
			val /= (1+mu*val1*(C1+C2+C3+C4)/(dx*dy)*dt);	// \varpi_{i,j}^{n+1}

			cvmSet(tmpmat, i,j, val);		
		}
	}
	cvCopy(tmpmat, varpi);
	//Re-initialization of the signed distance function
	ReInitDF(varpi);
	_time64( &timebuffer_after );
	_ftime64(&tstruct2);
	cvReleaseMat(&tmpmat);
	elapsetime_ms = (timebuffer_after-timebuffer_before)*1000+tstruct2.millitm-tstruct1.millitm;
	totol_elapsetime_ms += elapsetime_ms;

	iterate_num ++;
}
/*
	Calculate the EF via Graph cuts
*/
void ActWESolver::IterateGC()
{
	
	struct __timeb64 tstruct1,tstruct2;

	_time64( &timebuffer_before);
	_ftime64(&tstruct1);
	calC();	// calculate c1 and c2
	/*
		This threshold is set because the resolution of the image is only 1/255 > 0.001
		and c1, c2 is the average intensity value of the foreground and background respectively
	*/
	if((abs(prev_c1-c1)+abs(prev_c2-c2)) < 0.0001)
		return;
	else{
		prev_c1 = c1;
		prev_c2 = c2;
	}
	for(int j=0; j<width; j++){
		for(int i=0; i<height; i++){
			double val = cvmGet(u0,j,i);
			g -> set_tweights(i, j, lambda1*((c1-val)*(c1-val)), lambda2*((c2-val)*(c2-val)));
		}
	}
	// Maxflow
	g ->maxflow();
	for(int j=0; j<width; j++){
		for(int i=0; i<height; i++){
			if (g->what_segment(i, j) == TGraphImage::FOREGROUND){
				cvmSet(varpi,j,i,-1);
			}
			else{
				cvmSet(varpi,j,i,1);
			}
		}
	}
	_time64( &timebuffer_after );
	_ftime64(&tstruct2);
	elapsetime_ms = (timebuffer_after-timebuffer_before)*1000+tstruct2.millitm-tstruct1.millitm;
	totol_elapsetime_ms += elapsetime_ms;
	iterate_num ++;	
}

void ActWESolver::calC()
{
	double areac1 = 0, valc1 = 0;
	double areac2 = 0, valc2 = 0;
	double val;
	for(int i=0; i< width; i++)
		for(int j=0; j<height; j++){
			val = cvmGet(varpi, i,j ) ;
			if( val > 0){ //c1
				areac1 += 1;
				valc1 += cvmGet(u0, i,j);
			}else if( val < 0){ // c2
				areac2 += 1;
				valc2 += cvmGet(u0, i,j);
			}
		}

	c1 = valc1 / areac1;
	c2 = valc2 / areac2;
}

int ActWESolver::what_segment(int x, int y)
{
	if(cvmGet(varpi, x, y) >=0 ) 
		return 1;
	else
		return -1;
}

/*
	Calculate the denominator 
	sqrt(D_+^x \varpi_{i,j}^2)/dx^2 
	+ (\varpi_{i,j+1}-\varpi_{i,j-1})^2/(2*dx)^2) )}
*/
double ActWESolver::DenoX(int i, int j)
{
	double val, val0, val1;
	int x, y; //for saturation
	x = i; y = j+1;
	Saturate(x,y);
	val0 = cvmGet(varpi, x,y);
	x = i; y = j-1;
	Saturate(x,y);
	val0 -= cvmGet(varpi, x,y);	//(\varpi_{i,j+1}-\varpi_{i,j-1})
	val = Dx(1,i,j,1/2.0);			// U_+^x
	val1 = sqrt(val*val + val0 * val0 /(4*dy*dy)+MINIM); // The denominator
	return val1;
}

/*
	Calculate the denominator 
	sqrt(D_+^y \varpi_{i,j}^2)/dxdy
	+ (\varpi_{i+1,j}-\varpi_{i-1,j})^2/(2*dx)^2) )}
*/
double ActWESolver::DenoY(int i, int j)
{
	double val, val0, val1;
	int x, y; //for saturation
	x = i+1; y = j;
	Saturate(x,y);
	val0 = cvmGet(varpi, x,y);
	x = i-1; y = j;
	Saturate(x,y);
	val0 -= cvmGet(varpi, x,y);	//(\varpi_{i+1,j}-\varpi_{i-1,j})
	val = Dy(1,i,j,1/2.0);
	val1 = sqrt(val*val + val0 * val0 /(4*dx*dx) + MINIM); // The denominator

	return val1;
}

void ActWESolver::Extract_Contour(CvMat *disfun, CvMat *targetMat, double threshold)
{
	ASSERT((disfun->width == targetMat->width)&&(disfun->height == targetMat->height));

	cvZero(targetMat);
	double val0, val1, val_1, val_diag0;
	// Rows
	for(int i=1; i<(disfun->height-1); i++)
		for(int j=1; j<(disfun->width-1); j++){
			val0 = cvmGet(disfun, i, j);
			val1 = cvmGet(disfun, i+1,j);
			val_diag0 = cvmGet(disfun, i+1,j+1);
			if(val0 == 0){
				cvmSet(targetMat, i,j, 1);
				continue;
			}
			if( val0 * val1 < 0){
				if(val0>0){//fabs(val0) <= fabs(val1)
					cvmSet(targetMat, i,j, 1);	
				}else{
					cvmSet(targetMat, i+1,j, 1);	
				}	
//				val_1 = cvmGet(disfun, i-1,j);
//				if(val_1*val0 < 0){//An discontinuous node
//					if(!cvmGet(dev_rows,i,j)){
//						g->add_edge_weight(nodes[j*width+i],nodes[j*width+i-1] ,k/2,k/2);
//						g->add_edge_weight(nodes[j*width+i],nodes[j*width+i+1] ,k/2,k/2);
//						cvmSet(dev_rows,i,j,1);
//					}
//				}		
			}		
/*			if( val0 * val_diag0 < 0){//  Detect the 45 degree diagnal
				val_diag1 = cvmGet(disfun, i-1, j-1);
				if(val_diag1 * val0 <0){
					if(!cvmGet(dev_diag45,i,j)){
						g->add_edge_weight(nodes[j*width+i],nodes[(j+1)*width+i+1] ,k/2,k/2);
						g->add_edge_weight(nodes[j*width+i],nodes[(j-1)*width+i-1] ,k/2,k/2);
						cvmSet(dev_diag45,i,j,1);
					}
				}
			}
*/		}

		// Columns
		for(int i=1; i<(disfun->height-1); i++)
			for(int j=1; j<(disfun->width-1); j++){				
				val0 = cvmGet(disfun, i, j);
				val1 = cvmGet(disfun, i,j+1);
				val_diag0 = cvmGet(disfun, i+1,j-1);
				if(val0==0){
					cvmSet(targetMat, i,j, 1);
					continue;
				}

				if( val0 * val1 < 0){
					if(val0>0){//fabs(val0) <= fabs(val1)
						cvmSet(targetMat, i,j, 1);	
					}else{
						cvmSet(targetMat, i,j+1, 1);	
					}				
					val_1 = cvmGet(disfun, i,j-1);
//					if(val_1*val0 < 0){//An discontinuous node
//						if(!cvmGet(dev_cols,i,j)){
//							g->add_edge_weight(nodes[j*width+i],nodes[(j-1)*width+i] ,k/2,k/2);
//							g->add_edge_weight(nodes[j*width+i],nodes[(j+1)*width+i] ,k/2,k/2);
//							cvmSet(dev_cols,i,j,1);
//						}
//					}
				}
				
/*				if( val0 * val_diag0 < 0){//  Detect the 135 degree diagnal
					val_diag1 = cvmGet(disfun, i-1, j+1);
					if(val_diag1 * val0 <0){
						if(!cvmGet(dev_diag135,i,j)){
							g->add_edge_weight(nodes[j*width+i],nodes[(j-1)*width+i+1] ,k/2,k/2);
							g->add_edge_weight(nodes[j*width+i],nodes[(j+1)*width+i-1] ,k/2,k/2);
							cvmSet(dev_diag135,i,j,1);
						}
					}
				}
*/			}
}