#include "StdAfx.h"
#include ".\lssolver.h"
#include "../CVCommon.h"


LSSolver::LSSolver()
{
	loaded = false;
	varpi = NULL;
	ax = NULL;
	mux = NULL;
	varpi_x = NULL;
	varpi_y = NULL;
	convection_term = NULL;
	curvature_term = NULL;

	vout = NULL;
}
LSSolver::LSSolver( CvMat *varpi_i, CvMat *ax_i, CvMat *mux_i, 
				   double dt_i, bool video = false, char *videofname = "lsout.avi"):dt(dt_i)
{
	varpi = cvCloneMat(varpi_i);
	ax = cvCloneMat(ax_i);
	mux = cvCloneMat(mux_i);

	height = varpi->width; // Note this
	width = varpi->height;

	dx = 1.0;///width;
	dy = 1.0;///height;

	varpi_x = cvCloneMat(varpi_i);
	varpi_y = cvCloneMat(varpi_i);
	Gradient(varpi, varpi_x, varpi_y);

	convection_term = cvCreateMat(width, height, CV_32FC1);
	curvature_term = cvCreateMat(width, height, CV_32FC1);
	cvZero(convection_term);
	cvZero(curvature_term);

	if(video){
		vout= cvCreateVideoWriter(videofname, 541215044, 18, cvSize(width, height));
	}else{
		vout = NULL;
	}

	loaded = true;
}

LSSolver::~LSSolver(void)
{
	if(varpi)
		cvReleaseMat(&varpi);
	if(ax)
		cvReleaseMat(&ax);
	if(mux)
		cvReleaseMat(&mux);
	if(varpi_x)
		cvReleaseMat(&varpi_x);
	if(varpi_y)
		cvReleaseMat(&varpi_y);

	if(vout){
		cvReleaseVideoWriter(&vout);
	}
}

///////////////////////////////// Globally used functions //////////////////////////
void LSSolver::Saturate(int &i, int &j)
{
	if(i < 0) i = 0;
	if(j < 0) j = 0;
	if(i >= width) i = width -1;
	if(j >= height) j = height -1;
}

double LSSolver::M(double x, double y)
{
	if(fabs(x) <= fabs(y))
		return x;
	else 
		return y;
}


void LSSolver::Gradient(CvMat *I, CvMat *gradIx, CvMat *gradIy)
{
	int width = I->height;
	int height = I->width;
	double val;
	for(int i=0; i<width; i++)
		for(int j=0; j<height; j++){
			// For gradient x
			if(i>0 && i<(width-1)){
				val = (cvmGet(I, i+1, j) - cvmGet(I, i-1, j))/2.0;				
			}else if(i == 0){
				val = cvmGet(I, i+1, j) - cvmGet(I, i, j);
			}else{ // i==width -1
				val = cvmGet(I, i, j) - cvmGet(I, i-1, j);
			}
			cvmSet(gradIx, i, j, val);

			// For gradient y
			if(j>0 && j<(height-1)){
				val = (cvmGet(I, i, j+1) - cvmGet(I, i, j-1))/2.0;				
			}else if(j == 0){
				val = cvmGet(I, i, j+1) - cvmGet(I, i, j);
			}else{ // j==height -1
				val = cvmGet(I, i, j) - cvmGet(I, i, j-1);
			}
			cvmSet(gradIy, i, j, val);
		}
}
//////////////////////////////////////////////////////////////////////////

// Main iteration
void LSSolver::Iterate()
{
	if(!loaded)
		return;

	Gradient(varpi, varpi_x, varpi_y);
	convectionENO2_iterate();
	curvature_iterate();
	TVDRK2_iterate();

	ReInitDF(varpi);

}


/*
	Solve the reinitialization equation :
	\frac{\partial \psi}{\partial t} + sgn(\disfun(x))(|\nabla \psi|-1) = 0
	s.t.				\psi(x,0) = \disfun (x)
	
	We implement the slow version described in 
	"Active Contours without Edges"

	\psi_{i,j}^n+1 = \psi_{i,j}^n - \delta sign(\disfun(x)) G(\psi_{i,j}^n)
	where

	G(\psi_{i,j}) =  sqrt(max(a+^2, b-^2) + max(c+^2, d-^2)) -1         if \disfun(i,j) > 0
	                    =  sqrt(max(a-^2, b+^2) + max(c-^2, d+^2)) -1         if \disfun (i,j)  < 0
						= 0          otherwise
	and 
	a = (U_-^x), b = U_+^x, c=U_-^y, d=U_+^y
	x+ = max(x,0), x- = min(x,0)
*/
void LSSolver::ReInitDF(CvMat *disfun)
{
	CvMat *psi = cvCloneMat(disfun);
	CvMat *tmpMat = cvCloneMat(disfun);
	double err = 10;
	int numgrid = width * height;
	int max_iteration = 100;
	double dt = 0.01;
	double a, b, c, d;
	int sgnofdisfun;
	double val, val1;
	while(err > dt*dx*dx && max_iteration>0){
		err = 0;
		for(int i=1; i<width-1; i++){
			for(int j=1; j<height-1; j++){
				sgnofdisfun = sgn(cvmGet(disfun, i,j));
				a = Dx(psi, 1,i,j,-1/2.0);
				b = Dx(psi, 1,i,j,1/2.0);
				c = Dy(psi, 1,i,j,-1/2.0);
				d = Dy(psi, 1,i,j,1/2.0);
				switch(sgnofdisfun){
					case 1:
						val = dt * (sqrt(max(max(a,0)*max(a,0),min(b,0)*min(b,0)) 
							+ max(max(c,0)*max(c,0),min(d,0)*min(d,0))) - 1.0);
						break;
					case -1:
						val = dt * (sqrt(max(min(a,0)*min(a,0),max(b,0)*max(b,0)) 
							+ max(min(c,0)*min(c,0),max(d,0)*max(d,0))) - 1.0);
						break;
					default:
						val = 0;						
				}
				val1 = cvmGet(psi, i, j);
				if(fabs(val1) < dx*2)
					err += fabs(val);
				val = val1 - sgnofdisfun * val;
				cvmSet(tmpMat, i,j, val );
			}
		}// for 
		cvCopy(tmpMat, psi);
		err /= numgrid;
		max_iteration--;
	}//while
	
	cvCopy(psi, disfun);
	cvReleaseMat(&psi);
	cvReleaseMat(&tmpMat);
}

/*
	calculate the divided difference D_index^order+s \varpi of the X component
	s should be zero when order is an even number
	s can be 1/2 or -1/2 when order is an odd number
	the definition of D_index^order:
	D_i^0 \varpi = \varpi_i
	D_{i+1/2}^1 \varpi = \frac{D_{i+1}^0 - D_{i}^0}{\delta x}
	D_i^2 \varpi = \frac{D_{i+1/2}^1 - D_{1-1/2}^1}{2 \delta x}
*/
double LSSolver::Dx(int order, int indx, int indy, float s)
{
	if(order%2 == 0){// order is an even number
		ASSERT(s == 0);
		if(order>0){
			return (Dx(order-1,indx, indy, 1/2.0) - Dx(order-1,indx, indy, -1/2.0))/(order*dx) ;
		}else{// order = 0
			Saturate(indx, indy);
            return cvmGet(varpi, indx,indy);
		}
	}else{// order is an odd number
		ASSERT( s!= 0 && fabs(s)<1);
		if(s > 0){// s=1/2
			return (Dx(order-1, indx+1, indy, 0) - Dx(order-1, indx,indy,0))/(order*dx);
		}else{// s=-1/2
			return (Dx(order-1,indx,indy, 0) - Dx(order-1, indx-1, indy, 0))/(order*dx);
		}
	}
}

double LSSolver::Dy(int order, int indx, int indy, float s)
{
	if(order%2 == 0){// order is an even number
		ASSERT(s == 0);
		if(order>0){
			return (Dy(order-1,indx, indy, 1/2.0) - Dy(order-1,indx, indy, -1/2.0))/(order*dy) ;
		}else{// order = 0
			Saturate(indx, indy);
			return cvmGet(varpi, indx,indy);
		}
	}else{// order is an odd number
		ASSERT( s!= 0 && fabs(s)<1);
		if(s > 0){// s=1/2
			return (Dy(order-1, indx, indy+1, 0) - Dy(order-1, indx,indy,0))/(order*dy);
		}else{// s=-1/2
			return (Dy(order-1,indx,indy, 0) - Dy(order-1, indx, indy-1, 0))/(order*dy);
		}
	}
}

// A global form of Dx and Dy
double LSSolver::Dx(CvMat* mat, int order, int indx, int indy, float s)
{
	if(order%2 == 0){// order is an even number
		ASSERT(s == 0);
		if(order>0){
			return (Dx(mat, order-1,indx, indy, 1/2.0) - Dx(mat, order-1,indx, indy, -1/2.0))/(order*dx) ;
		}else{// order = 0
			Saturate(indx, indy);
			return cvmGet(mat, indx,indy);
		}
	}else{// order is an odd number
		ASSERT( s!= 0 && fabs(s)<1);
		if(s > 0){// s=1/2
			return (Dx(mat, order-1, indx+1, indy, 0) - Dx(mat, order-1, indx,indy,0))/(order*dx);
		}else{// s=-1/2
			return (Dx(mat, order-1,indx,indy, 0) - Dx(mat, order-1, indx-1, indy, 0))/(order*dx);
		}
	}
}

double LSSolver::Dy(CvMat*mat, int order, int indx, int indy, float s)
{
	if(order%2 == 0){// order is an even number
		ASSERT(s == 0);
		if(order>0){
			return (Dy(mat, order-1,indx, indy, 1/2.0) - Dy(mat, order-1,indx, indy, -1/2.0))/(order*dy) ;
		}else{// order = 0
			Saturate(indx, indy);
			return cvmGet(mat, indx,indy);
		}
	}else{// order is an odd number
		ASSERT( s!= 0 && fabs(s)<1);
		if(s > 0){// s=1/2
			return (Dy(mat, order-1, indx, indy+1, 0) - Dy(mat, order-1, indx,indy,0))/(order*dy);
		}else{// s=-1/2
			return (Dy(mat, order-1,indx,indy, 0) - Dy(mat, order-1, indx, indy-1, 0))/(order*dy);
		}
	}
}


/*
	Calculation of the convection term using ENO  order 2 method

	a(x).\nabla\varpi = (a1(xj,yk,tn)^+) D_-^x \varpi_jk 
								+ (a1(xj,yk,tn)^-) D_+^x \varpi_jk
								+ (a2(xj,yk,tn)^+)D_-^y \varpi_jk
								+ (a2(xj,yk,tn)^-) D_+^y\varpi_jk
	where x^+ = max(x,0) and x^- = min(x,0);

ENO method:
	I_{j-1/2,k}(x) = \varpi_{j,k} + (x-x_j) D_-^x \varpi_jk
	which  d/dx (I_(j-1/2)^1 (x) )|_{x=x_j} = D_-^x \varpi_jk

	for order 2 accuracy
	I_{j-1/2,k}^2(x) = I_{j-1,k}^1(x) + (x-x_j)(x-x_{j-1})/2 M(D_-^2 /varpi_jk, D_-D_+ /varpi_jk)

Thus the mth order ENO approximation to a(x).\nabla \varpi is:
		(a1(xj,yk,tn)^+) (d/dx (I_{j-1/2,k}^m(x)))_x=xj
		+ (a1(xj,yk,tn)^-) (d/dx (I_{j+1/2,k}^m(x)))_x=xj
		+ (a2(xj,yk,tn)^+) (d/dy (I_{j,k-1/2}^m(x)))_y=yk
		+ (a2(xj,yk,tn)^-) (d/dy (I_{j,k+1/2}^m(x)))_y=yk
when m=2
	d/dx (I_{j-1/2,k}^2(x) = D_-^x \varpi_jk + (dx) M(D_j-1^2 \varpi_jk, D_j^2 /varpi_jk)
	d/dx (I_{j+1/2,k}^2(x) = D_+^x \varpi_jk - (dx) M(D_j^2 \varpi_jk, D_j+1^2 /varpi_jk)
	d/dy (I_{j,k-1/2}^2(x) = D_-^y \varpi_jk + (dy) M(D_k-1^2 \varpi_jk, D_k^2 /varpi_jk)
	d/dy (I_{j,k+1/2}^2(x) = D_+^y \varpi_jk - (dy) M(D_k^2 \varpi_jk, D_k+1^2 /varpi_jk)
*/

void LSSolver::convectionENO2_iterate()
{
	double val;
	for(int i=0; i<width; i++){
		for(int j=0; j<height; j++){
			double a1,a2, varpix, varpiy, sqvarxy;
			val = cvmGet(ax,i,j);
			varpix = cvmGet(varpi_x, i,j);		// \varpi_x(i,j)
			varpiy = cvmGet(varpi_y, i,j);		// \varpi_y(i,j)
			sqvarxy = sqrt(varpix*varpix + varpiy*varpiy + MINIM);// |sqrt(\varpi_x^2 + \varpi_y^2)|
			a1 = val * varpix / sqvarxy;
			a2 = val * varpiy / sqvarxy;
			
			val = max(a1, 0)  * (Dx(1,i,j,-1/2.0) + dx *M(Dx(2, i-1,j,0), Dx(2,i,j,0)))
					+ min(a1,0) * (Dx(1,i,j,1/2.0) - dx * M(Dx(2,i+1,j,0), Dx(2,i,j,0)))	
					+ max(a2,0) * (Dy(1,i,j,-1/2.0) + dy * M(Dy(2,i,j-1,0), Dy(2,i,j,0)))
					+ min(a2,0) * (Dy(1,i,j,1/2.0) - dy * M(Dy(2,i,j+1,0), Dy(2,i,j,0)));

			cvmSet(convection_term, i,j,val);
		}
	}
}

/*
	Calculation of the curvature term mu(x) |/delta /varpi| /delta.(/frac{/delta/varpi}{|/delta/varpi|})	
	where
	|/delta /varpi| = sqrt((D_+^x + D_-^x + D_+^y + D_-^y)/2)
	/delta . (\frac{/delta /varpi}{|/delta/varpi|}) = /partial / (/partial x) (/frac{/varpi_x}{|/delta/varpi|})
																		+ /partial / (/partial y) (/frac{/varpi_y}{|/delta/varpi|})
	where
	/partial / (/partial x) (/frac{/varpi_x}{|/delta/varpi|}) 
	= D_-^x (/frac{D_+^x /varpi_jk}{sqrt((D_+^x/varpi_jk)^2+(D^y /varpi_jk)^2+MINIM)})
	and
	(D^y /varpi_jk)^2 = ((D_+^y/varpi_jk)^2+(D_+^y/varpi_j+1,k)^2
								+(D_-^y/varpi_jk)^2+(D_-^y(/varpi_j+1,k))^2)/4
	
*/
void LSSolver::curvature_iterate()
{
	double val;
	for(int i=0; i<width; i++)
		for(int j=0; j<height; j++){
			// the |\delta \varpi| term
			double delta_varpi = sqrt((Dx(1,i,j,1/2.0)*Dx(1,i,j,1/2.0) 
											+ Dx(1,i,j,-1/2.0)*Dx(1,i,j,-1/2.0)
											+ Dy(1,i,j, 1/2.0) *Dy(1,i,j,1/2.0)
											+ Dy(1,i,j,-1/2.0)*Dy(1,i,j,-1/2.0))/2.0);
			// the (D^y \varpi_jk)^2 term
			double Dyvarpi_jk = ( Dy(1,i,j,1/2.0)*Dy(1,i,j,1/2.0)
										+Dy(1,i+1,j,1/2.0)*Dy(1,i+1,j,1/2.0)
										+Dy(1,i,j,-1/2.0)*Dy(1,i,j,-1/2.0)
										+Dy(1,i+1,j,-1/2.0)*Dy(1,i+1,j,-1/2.0))/4;
			double Dyvarpi_jm1k = ( Dy(1,i-1,j,1/2.0)*Dy(1,i-1,j,1/2.0)
										+Dy(1,i,j,1/2.0)*Dy(1,i,j,1/2.0)
										+Dy(1,i-1,j,-1/2.0)*Dy(1,i-1,j,-1/2.0)
										+Dy(1,i,j,-1/2.0)*Dy(1,i,j,-1/2.0))/4;
			// (/partial / /partial x) /frac{/varpi_x}{|/delta /varpi|}
			double partial_x_varpi = Dx(1,i,j,1/2.0) / sqrt(Dx(1,i,j,1/2.0)*Dx(1,i,j,1/2.0) + Dyvarpi_jk + MINIM)
										- Dx(1,i-1,j,1/2.0) / sqrt(Dx(1,i-1,j,1/2.0)*Dx(1,i-1,j,1/2.0) + Dyvarpi_jm1k + MINIM);
			
			// the (D^x \varpi_jk)^2 term
			double Dxvarpi_jk = ( Dx(1,i,j,1/2.0)*Dy(1,i,j,1/2.0)
				+Dx(1,i,j+1,1/2.0)*Dx(1,i,j+1,1/2.0)
				+Dx(1,i,j,-1/2.0)*Dx(1,i,j,-1/2.0)
				+Dx(1,i,j+1,-1/2.0)*Dx(1,i,j+1,-1/2.0))/4.0;
			double Dxvarpi_jkm1 = ( Dx(1,i,j-1,1/2.0)*Dx(1,i,j-1,1/2.0)
				+Dx(1,i,j,1/2.0)*Dx(1,i,j,1/2.0)
				+Dx(1,i,j-1,-1/2.0)*Dx(1,i,j-1,-1/2.0)
				+Dx(1,i,j,-1/2.0)*Dx(1,i,j,-1/2.0))/4.0;
			// (\partial / \partial y) \frac{\varpi_y}{|\delta \varpi|}
			double partial_y_varpi = Dy(1,i,j,1/2.0) / sqrt(Dy(1,i,j,1/2.0)*Dy(1,i,j,1/2.0) + Dxvarpi_jk + MINIM)
										- Dy(1,i-1,j,1/2.0) / sqrt(Dy(1,i-1,j,1/2.0)*Dy(1,i-1,j,1/2.0) + Dxvarpi_jkm1 + MINIM);
			val = cvmGet(mux,i,j);	//mu(x)_ij
			val = val * delta_varpi * (partial_x_varpi + partial_y_varpi);
			cvmSet(curvature_term, i,j, val);
		}
}

/*
	calculate the time discretization term  
	frac{/partial /varpi}{/partial t} = curvture_term - convection_term
	using 2nd order TVD Runge-Kutta method	
*/
void LSSolver::TVDRK2_iterate()
{
	CvMat *tmp = cvCreateMat(width, height, CV_32FC1);
	double val;
	for(int i = 0; i<width; i++)	// the \varpi_jk^{n+1/2} term
		for(int j=0; j<height; j++){
			val = cvmGet(varpi, i,j) + dt * ( cvmGet( curvature_term, i,j) - cvmGet( convection_term, i,j));// +cvmGet( curvature_term, i,j)
			cvmSet(tmp, i,j, val);
		}

	for(int i = 0; i<width; i++)
		for(int j=0; j<height; j++){						
			val = cvmGet(tmp, i,j);//+ dt * (cvmGet( convection_term, i,j) - cvmGet( curvature_term, i,j));
			cvmSet(varpi, i,j, val);
		}

	cvReleaseMat(&tmp);
}

/*
		Extract the interface
*/
void LSSolver::Extract_Contour(CvMat *disfun, CvMat *targetMat, double threshold=0.1)
{
	ASSERT((disfun->width == targetMat->width)&&(disfun->height == targetMat->height));

	cvZero(targetMat);
	double val0, val1;
	// Rows
	for(int i=0; i<(disfun->height-1); i++)
		for(int j=0; j<disfun->width; j++){
			val0 = cvmGet(disfun, i, j);
			val1 = cvmGet(disfun, i+1,j);
			if(val0 == 0){
				cvmSet(targetMat, i,j, 1);
				continue;
			}
			if( val0 * val1 < 0){
				if(val0>0){//fabs(val0) <= fabs(val1)
					cvmSet(targetMat, i,j, 1);	
				}else{
					cvmSet(targetMat, i+1,j, 1);	
				}				
			}
		}

	// Columns
	for(int i=0; i<(disfun->height); i++)
		for(int j=0; j<(disfun->width-1); j++){
			val0 = cvmGet(disfun, i, j);
			val1 = cvmGet(disfun, i,j+1);
			if(val0==0){
				cvmSet(targetMat, i,j, 1);
				continue;
			}
			
			if( val0 * val1 < 0){
				if(val0>0){//fabs(val0) <= fabs(val1)
					cvmSet(targetMat, i,j, 1);	
				}else{
					cvmSet(targetMat, i,j+1, 1);	
				}				
			}
		}
}

CvMat *LSSolver::GetResult()
{
	return  varpi;
}

void LSSolver::output2movie(IplImage *frame)
{
	if(vout){		
		cvWriteFrame(vout,frame);		
	}
}