/* graph.h */
/* Vladimir Kolmogorov (vnk@cs.cornell.edu), 2001. */
/* Modified Version by Xiang Zeng (iamzengxiang@gmail.com) 2006*/

/*	
	Class Graph defines basic structures and operations in a flow network.
*/

#pragma once

#include "block.h"

#include <crtdbg.h>

class Graph
{
///////////////////// Interfaces ////////////////////
public:
	/* optional for debugging */
	typedef void (*ErrorFunction)(char* errmsg);

//////// public variables /////////
public:
	/* S tree = SOURCE tree
		T tree = SINK tree */
	enum TreeType { SOURCE, SINK, NONE };
	/* If a node belongs to a  S or T tree, 
		its state must be either ACTIVE or PASSIVE,
		else its state is FREE	*/
	enum NodeState { FREE, ACTIVE, PASSIVE };
	NodeState STATE_FREE; // = FREE for initialize and 

	/* Type of edge weights */
	typedef double CapType;
	/* Type of total flow */
	typedef int FlowType;

private: /* this is to prevent the public method of Graph to be called */
	/* Pointer to the node struct */
	typedef void * node_id;

/////// public functions ///////// 
public:
	/* Adds a node to the graph */
	node_id add_node();

	/* Add a bidirectional edge between 'from' and 'to'
	with the weights 'cap' and 'rev_cap' */
	void add_edge(node_id from, node_id to, CapType cap, CapType rev_cap);

	/* Set the weights of the edges 'SOURCE->i' and 'i->SINK'
	Can be called at most once for each node before any call to 'add_tweights'.
	Weights can be negative */
	void set_tweights(node_id i, CapType cap_source, CapType cap_sink);

	/* Add new edges 'SOURCE->i' and 'i->SINK' with corresponding weights
	Can be called multiple times for each node.
	Weights can be negative */
	void add_tweights(node_id i, CapType cap_source, CapType cap_sink);

	/* After the maxflow is computed, this function returns to which
	segment the node 'i' belongs (Graph::SOURCE or Graph::SINK) */
	TreeType what_segment(node_id i);

	/* Computes the maxflow. Can be called only once. */
	virtual FlowType maxflow() = 0;

public:
	Graph(void (*ErrorFunction)(char *) = NULL);
	~Graph(void);






////////////////// These definition can be changed by inheritance ////////////////////
///////// protected constants and structures /////////
protected:
	const int NODE_BLOCK_SIZE;
	const int ARC_BLOCK_SIZE;
	const int NODEPTR_BLOCK_SIZE;

	const int INFINITE_D;

	struct arc;

	/* node structure */
	struct node 
	{
		int					DIST;				/* distance to the boundary */

		TreeType		tree_type;		/* flag showing whether the node belongs to the source(S) or sink(T) tree, or neither*/

		NodeState		*node_state;		/* If tree_type == NONE  then state = FREE
														else								state = ACTIVE	or PASSIVE 
														making it a pointer is for fast state switching */

		CapType		tr_cap;				/* if tr_cap > 0 then  tr_cap is residual capacity of the arc SOURCE->node
														otherwise				-tr_cap is residual capacity of the arc node->SINK */
	};

	/* arc structure */
	struct arc
	{
		arc			*sister;			/* reverse arc */

		CapType		r_cap;			/* residual capacity */
	};


///////// protected variables /////////////
protected:
	ErrorFunction error_function;

	FlowType			flow;			/* total flow */

	Block<node>			*node_block;
	Block<arc>			*arc_block;

////////// protected functions ////////////
protected:

	//////// To be implemented by inheritance ////////
	virtual void maxflow_init() = 0;

	void augment(arc *middle_arc); /* no implementation */

	void process_source_orphan(node *i); /* no implementation */
	void process_sink_orphan(node *i); /* no implementation */

};
