/*
	Layered Priority Queue (LPQueue)
	A companion Class for fast popping and pushing nodes in TGraphImage
	It is memory and speed efficient.

	Its difference with the normal Priority Queue is that
	the current priority is given to those nodes with the 
	current processed key value (cur_key), instead of 
	the minimal key value (min_key). And until there is 
	no node with the key value cur_key, the current key
	value is set to the 	minimal key value (min_key).
*/

#pragma once

#include <crtdbg.h>

template<class Item_Type, class Tag_Type>
class LPQueue
{
	///// This class can only be used by TGraphImage class /////
	friend class TGraphImage;


private: //////// private variables ///////////
	struct array_node
	{
		Item_Type *head;
		Item_Type *tail;
		array_node(): head(NULL), tail(NULL) {}
	};
	
	Tag_Type tag;

	array_node *arr_ptr;
	
	const int LOWER_BOUND, UPPER_BOUND;
	int item_count;
	int current_index, min_index;
	
private: //////// private functions ///////////
	/* push a new item into the array, O(1) complexity*/
    void push_back(Item_Type *_item)
	{
		int idx = _item->node_ptr->DIST - LOWER_BOUND;

		if(arr_ptr[idx].head == NULL) { // empty array
			arr_ptr[idx].head = arr_ptr[idx].tail = _item;
			_item->q_next = _item->q_prev = NULL;
		}else{ // add item pointer to the tail
			arr_ptr[idx].tail->q_next = _item;
			_item->q_prev = arr_ptr[idx].tail;
			_item->q_next = NULL;
			arr_ptr[idx].tail = _item;
		}

		if( idx < min_index ) {
			min_index = idx;
			if(arr_ptr[current_index].head == NULL) {
				current_index = min_index;
			}
		}

		_item->node_ptr->node_state = &tag;

		item_count ++;
	}

	/* O(1) complexity */
	void delete_item(Item_Type *_item)
	{
		if(!item_count) return;

		int idx = _item->node_ptr->DIST - LOWER_BOUND;

		if(arr_ptr[idx].head == _item)
			arr_ptr[idx].head = _item->q_next;
		if(arr_ptr[idx].tail == _item)
			arr_ptr[idx].tail = _item->q_prev;

		if(_item->q_prev != NULL) 
			_item->q_prev->q_next = _item->q_next;
		if(_item->q_next != NULL)
			_item->q_next->q_prev = _item->q_prev;

		_item->q_prev = _item->q_next = NULL;

		item_count --;

		if(item_count == 0) {// reset
			min_index = UPPER_BOUND - LOWER_BOUND + 1;
			current_index = 0;
			return;
		}else if( arr_ptr[min_index].head == NULL ) {// empty in min_index
			while(++min_index <= (UPPER_BOUND - LOWER_BOUND)) 
				if(arr_ptr[min_index].head) {					
					break;					
				}
		}

		/* make sure current_index always point to a non-empty  item*/
		if( arr_ptr[current_index].head == NULL) { // item_count > 0
			if(current_index != min_index) {
				current_index = min_index;
				if(min_index > (UPPER_BOUND - LOWER_BOUND)) {
					current_index = min_index  = 0;
					item_count = 0;
				}
			}else {
//  				while(++current_index <= (UPPER_BOUND - LOWER_BOUND)) 
//  					if(arr_ptr[current_index].head) {
//  						min_index = current_index;
//  						break;					
//  					}
//  				if(current_index > (UPPER_BOUND - LOWER_BOUND)) 
// 					_ASSERT(false);
			}
		}

	}

	/* O(1) complexity */
	Item_Type *extract_front()
	{
		Item_Type *tmp = arr_ptr[current_index].head;
		pop();

		if(tmp != NULL && arr_ptr[current_index].head == NULL && item_count > 0) {
			if(current_index != min_index) 
				current_index = min_index;
			else {
				while(++current_index <= (UPPER_BOUND - LOWER_BOUND)) 
					if(arr_ptr[current_index].head) {
						min_index = current_index;
						break;				
					}
			}
		}		
		return tmp;
	}

	void begin() 
	{
		if(item_count > 0) 
			current_index = min_index;
	}

	bool is_empty() { return (item_count == 0); }

	Item_Type* front() { return arr_ptr[current_index].head;	}

	void pop() { delete_item(arr_ptr[current_index].head); }

	void set_tag(Tag_Type _tag) { tag = _tag; }

	Tag_Type get_tag() { return tag; }

public:
	LPQueue(int min_val, int max_val, Tag_Type tag_val):
	  LOWER_BOUND(min_val), UPPER_BOUND(max_val), item_count(0), tag(tag_val),
		  current_index(0), min_index(UPPER_BOUND - LOWER_BOUND + 1)
	{		
		_ASSERT(LOWER_BOUND < UPPER_BOUND);
		arr_ptr = new array_node[UPPER_BOUND - LOWER_BOUND + 1];		
	};

	~LPQueue(void) { delete[] arr_ptr; }
};
