/*
	The topology preserving Max-Flow/Min-Cut algorithm : top-cut
*/

/*
Image term clarification:

			|---height-----|              
	---	---------------> y direction
	 |		|    
	 |		|    
   width |  
	 |		|    
	 |		|    
	---   \/ x direction

----------------The direction definition---------------

							up: (x-1, y) 0
								|
								|
left: (x, y-1) 2----(x,y)---- right: (x, y+1) 1
								|
								|
							down: (x+1, y) 3
*/

#pragma once

#include <queue>

#include "graph.h"
#include "SDistMap.h"
#include "LPQueue.h"

using namespace std;

class TGraphImage :
	public Graph
{
/////////////////////////// Interface //////////////////////////////
//////////// public variables //////////
public:	
	/*	label indicating the topology property of a node */
	enum TLabel {BACKGROUND, FOREGROUND};
	/* 4- or 8- neighborhood system */
	enum NBType {NB_FOUR, NB_EIGHT};

/////////// public functions //////////
public:
	/* add a new closed contour by giving a few key points 
		currently no intersection is allowed. */
	void add_contour(int *x, int *y, int numOfPnts);

	void add_edge(int x1, int y1, int x2, int y2, CapType cap);

	void set_tweights(int x, int y, CapType cap_source, CapType cap_sink);

	TLabel what_segment(int x, int y);

	FlowType maxflow();

public:
	TGraphImage(int width, int height, ErrorFunction errorFun = NULL);
	~TGraphImage(void);



//////////////////////////// Implementation ///////////////////////
/////// constants and structures //////////
private:
	/* Additional information recoding the topology label and coordinate for each node */
	struct t_node
	{
		node		*node_ptr;

		arc		*arc_ptr[4]; // Up, Right, Left, Down

		/* use parent_node for tracking feasible flow path */
		t_node *parent_node;	/* the parent node in the SOURCE or SINK tree  */
		int			parc_ind;			/* the index of the arc to the parent */

		TLabel t_label;				/* topology label */

		int pos_x, pos_y;			/* coordinate in image grid */

		t_node* q_next;			/* for fast active/passive set operation */
		t_node* q_prev;
	};

	/* special constants for node->parent */
#define TERMINAL ( (t_node *) 1 )		/* to terminal */
#define ORPHAN   ( (t_node *) 2 )		/* orphan */

	// The min heap for the active and passive sets
	typedef LPQueue<t_node, Graph::NodeState> TNODE_LPQUEUE;
	typedef queue<t_node*> TNODE_QUEUE;

/////// variables ///////
private:
	const int m_width, m_height;
	
	/* type of the neighborhood connectivity for foreground object, 8-connectivity is smoother */
	NBType nb_type;

	TNODE_LPQUEUE *act_Sf, *act_Sb, *act_Tf, *act_Tb;
	TNODE_LPQUEUE *pas_Sf, *pas_Sb, *pas_Tf, *pas_Tb;
	TNODE_QUEUE orphan_set;

	TNODE_QUEUE foreground_simple_set, background_simple_set; // record those nodes whose label can be changed 


	/* store the nodes in image grid, size = width * height */
	t_node *nodes;

	/* the topology label and  the distance to boundary of each are both calculated by SDistMap
		the relationship between Tlabel and SDistMap::Label is 
		BACKGROUND = SDistMap::OUTER	
		FOREGROUND = SDistMap::EDGE + SDistMap::INNER */
	SDistMap *sdmap;

/////// functions //////
private:
	////// utility functions //////
	inline t_node* getNode(int x, int y) { return &nodes[x*m_height + y]; }
	inline TLabel getTLabel(t_node *nd) { return nd->t_label; }

#define IN_IMAGE(x, y)		(x >= 0 && x < m_width && y>=0 && y < m_height)
#define ND_STATE(nd)			*(nd->node_ptr->node_state)

/* index in neighboring arc of structure node */
#define ARC_UP			0
#define ARC_RIGHT		1
#define ARC_LEFT			2
#define ARC_DOWN		3

	/* if a node in (x, y) is simple */
	bool is_simple(int x, int y);

	/* extract the next active node in the min_heap act_set */
	t_node* next_active(TNODE_LPQUEUE *act_set);
	/* extract the next active node with the value max {act_set1.top(), act_set2.top()}
	   -- the one nearest to the boundary first*/
	t_node* next_active(TNODE_LPQUEUE *act_set1, TNODE_LPQUEUE *act_set2);
	void set_active(TNODE_LPQUEUE *act_set, t_node *nd);
	void set_passive(TNODE_LPQUEUE *pas_set, t_node *nd);

	/* make node2 not free (a terminal) while preserving the original label 
        when node1 is a source node, arc_ptr = (node1, node2) 
		when node1 is a sink node, arc_ptr = (node2, node1) */
	void transform(t_node *node1, t_node *node2, arc *arc_ptr, bool is_sink);

	void exchange_set(TNODE_LPQUEUE *&set1, TNODE_LPQUEUE *&set2);
	void empty_set(TNODE_LPQUEUE *set);
	void delete_node(TNODE_LPQUEUE *set, t_node *nd);


	//////// main functions that implements the algorithm //////
	void maxflow_init();

	/* topology preserving Max-flow*/
	FlowType tp_maxflow();

	arc* source_boundary_competition_flow(
		t_node *n_src,							// currently processed node of SOURCE tree
		t_node* &n_output,					// if a feasible path is found, n_output stores the other node
		TNODE_LPQUEUE *act_Si,			// the active set where n_src belongs to
		TNODE_LPQUEUE *act_So,		// the active set with the different label from n_src
		TNODE_LPQUEUE *pas_So		// the passive set with the different label from n_src
		);	
	arc* sink_boundary_competition_flow(
		t_node *n_sink,						// currently processed node of SINK tree
		t_node* &n_output,					// if a feasible path is found, n_output stores the other node
		TNODE_LPQUEUE *act_Ti,			// the active set where n_sink belongs to
		TNODE_LPQUEUE *act_To,		// the active set with the different label from n_sink
		TNODE_LPQUEUE *pas_To		// the passive set with the different label from n_sink
		);		
	arc* foreground_intra_region_flow(
		t_node *n_sink,						// currently processed node in Act_Tf 
		t_node* &n_output					// if a feasible path is found, n_output stores the other node
		);
	arc* background_intra_region_flow(
		t_node *n_src,							// currently processed node n Act_Sb
		t_node* &n_output					// if a feasible path is found, n_output stores the other node
		);

	void augment(t_node *src_nd, t_node *sink_nd, arc *arc_middle);

	/* act_Si, the active set that nd 'may' belong to */
	void process_source_orphan(t_node *nd);
	void process_sink_orphan(t_node *nd);
};